import copy
import datetime
import hashlib
import functools
import os
import re
import time
import urlparse

import appdirs
import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref


def switch_to_dir(dirPath):
    """
    Decorator that switches to given directory before executing function, and
    then returning to orignal directory.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            orig_cwd = os.getcwd()
            os.chdir(dirPath)
            ret = func(*args, **kwargs)
            os.chdir(orig_cwd)
            return ret
        return wrapper

    return decorator


def _cacheValid_pfr(ct, mt, fn):
    # TODO: this is not comprehensive, can be a problem
    # first, if we can ensure that the file won't change,
    # then we're safe caching it
    import ipdb
    ipdb.set_trace()
    if 'boxscore' in fn:
        return True
    # if it's currently the offseason, then we're good
    today = datetime.date.today()
    endOfSeason = datetime.date(today.year, 2, 18)
    startOfSeason = datetime.date(today.year, 8, 25)
    if endOfSeason < today < startOfSeason:
        return True
    # now, check for a year in the filename
    m = re.search(r'(\d{4})', fn)
    if not m:
        if 'teams' in fn:
            return True
        else:
            return False
    year = int(m.group(1))
    # if it was a year prior to the current season, we're good
    curSeason = today.year - (today <= endOfSeason)
    if year < curSeason:
        return True
    # otherwise, check if new data could have been updated since mod
    # (assumed that new game data is added the day after that game)
    modDay = today - datetime.timedelta(seconds=ct - mt)
    lastGameDay = today
    while (lastGameDay.weekday() + 1) % 7 not in (0, 1, 2, 5):
        lastGameDay = lastGameDay - datetime.timedelta(days=1)
    return modDay >= lastGameDay


def _cacheValid_bkref(ct, mt, fn):
    # TODO: this might not be comprehensive
    # first, if we can ensure that the file won't change,
    # then we're safe caching it
    if 'boxscore' in fn:
        return True
    # now, check for a year in the filename
    m = re.search(r'(\d{4})', fn)
    if not m:
        return False
    year = int(m.group(1))
    today = datetime.date.today()
    endOfSeason = datetime.date(today.year, 6, 30)
    startOfSeason = datetime.date(today.year, 9, 23)
    # if it was a year prior to the current season, we're good
    curSeason = today.year - (today <= endOfSeason)
    if year < curSeason:
        return True
    # if we're in the offseason, we're good
    if endOfSeason < today < startOfSeason:
        return True
    # otherwise, check if new data could have been updated since mod
    # (assumed that new game data is added the day after that game)
    modDay = today - datetime.timedelta(seconds=ct - mt)
    return modDay >= today


def _cacheValid_cfb(ct, mt, fn):
    # TODO: caching for CFB
    return True


def cache_html(func):
    """Caches the HTML returned by the specified function `func`. Caches it in
    the user cache determined by the appdirs package.
    """

    CACHE_DIR = appdirs.user_cache_dir('sportsref', 'mgoldberg')
    if not os.path.isdir(CACHE_DIR):
        os.makedirs(CACHE_DIR)

    def cacheValidFuncs(s):
        return eval('_cacheValid_' + s)

    @functools.wraps(func)
    def wrapper(url):
        parsed = urlparse.urlparse(url)
        sport = sportsref.SITE_ABBREV.get(parsed.scheme + '://' +
                                          parsed.netloc)
        if sport is None:
            for ncaaSport in ('cfb', 'cbb'):
                if ncaaSport in url:
                    sport = ncaaSport
        relURL = parsed.path
        if parsed.query:
            relURL += '?' + parsed.query
        noPathFN = re.sub(r'\.html?', '', sport + relURL.replace('/', ''))
        file_hash = hashlib.md5()
        file_hash.update(noPathFN)
        file_hash = file_hash.hexdigest()
        filename = '{}/{}'.format(CACHE_DIR, file_hash)

        # set time variables (in seconds)
        if os.path.isfile(filename):
            curtime = int(time.time())
            modtime = int(os.path.getmtime(filename))
            time_since_mod = datetime.timedelta(seconds=(curtime - modtime))
            cacheValid = time_since_mod <= datetime.timedelta(days=12)

        # if file found and caching is valid, read from file
        if os.path.isfile(filename) and cacheValid:
            with open(filename, 'r') as f:
                text = f.read()
            return text
        # otherwise, download html and cache it
        else:
            text = func(url)
            with open(filename, 'w+') as f:
                f.write(text.encode('ascii', 'replace'))
            return text

    return wrapper


def memoized(fun):
    """A simple memoize decorator."""
    @functools.wraps(fun)
    def wrapper(*args, **kwargs):

        # deal with lists in args
        def isList(a):
            return isinstance(a, list) or isinstance(a, np.ndarray)

        def deListify(arg):
            if isList(arg):
                return tuple(map(deListify, arg))
            else:
                return arg

        # deal with dicts in args
        def isDict(d):
            return isinstance(d, dict) or isinstance(d, pd.Series)

        def deDictify(arg):
            if isDict(arg):
                items = dict(arg).items()
                items = [(k, deListify(deDictify(v))) for k, v in items]
                return frozenset(sorted(items))
            else:
                return arg

        clean_args = tuple(map(deListify, args))
        clean_args = tuple(map(deDictify, clean_args))
        clean_kwargs = deDictify(kwargs)

        def _copy(v):
            if isinstance(v, pq):
                return v.clone()
            else:
                return copy.deepcopy(v)

        key = (clean_args, clean_kwargs)
        try:
            ret = _copy(cache[key])
            return ret
        except KeyError:
            cache[key] = fun(*args, **kwargs)
            ret = _copy(cache[key])
            return ret
        except TypeError:
            print 'memoization type error here', fun.__name__, key
            return fun(*args, **kwargs)

    cache = {}
    return wrapper


def kind_rpb(include_type=False):
    def decorator(fun):
        """Supports functions that return a DataFrame and have a `kind` keyword
        argument that specifies regular season ('R'), playoffs ('P'), or both
        ('B'). If given 'B', it will call the function with both 'R' and 'P'
        and concatenate the results.
        """
        @functools.wraps(fun)
        def wrapper(*args, **kwargs):
            kind = kwargs.get('kind', 'R').upper()
            if kind == 'B':
                kwargs['kind'] = 'R'
                reg = fun(*args, **kwargs)
                if include_type:
                    reg['is_playoffs'] = False
                kwargs['kind'] = 'P'
                poffs = fun(*args, **kwargs)
                if include_type:
                    poffs['is_playoffs'] = True
                return pd.concat((reg, poffs), ignore_index=True)
            else:
                df = fun(*args, **kwargs)
                if include_type:
                    df['is_playoffs'] = (kind == 'P')
                return df
        return wrapper
    return decorator



# flake8: noqa

SITE_ABBREV = {
    'http://www.pro-football-reference.com': 'pfr',
    'http://www.basketball-reference.com': 'bkref',
    'http://www.sports-reference.com/cfb': 'ncaaf',
    'http://www.sports-reference.com/cbb': 'ncaab',
}

import decorators
import utils
import nfl
import nba

__all__ = ['decorators', 'utils', 'nfl', 'nba', 'SITE_ABBREV']



import numpy as np
from scipy.stats import norm


def initialWinProb(line):
    """Gets the initial win probability of a game given its Vegas line.

    :line: The Vegas line from the home team's perspective (negative means
    home team is favored).
    :returns: A float in [0., 100.] that represents the win probability.
    """
    line = float(line)
    probWin = 1. - norm.cdf(0.5, -line, 13.86)
    probTie = norm.cdf(0.5, -line, 13.86) - norm.cdf(-0.5, -line, 13.86)
    return 100. * (probWin + 0.5 * probTie)


def winProb(line, margin, secsElapsed, expPts):
    line = float(line)
    margin = float(margin)
    expPts = float(expPts)
    baseMean = -line
    baseStd = 13.46
    expMargin = margin + expPts
    minRemain = 60. - secsElapsed / 60. + 0.00001
    adjMean = baseMean * (minRemain / 60.)
    adjStd = baseStd / np.sqrt(60. / minRemain)
    probWin = 1. - norm.cdf(-expMargin + 0.5, adjMean, adjStd)
    probTie = (norm.cdf(-expMargin + 0.5, adjMean, adjStd) -
               norm.cdf(-expMargin - 0.5, adjMean, adjStd))
    return 100. * (probWin + 0.5 * probTie)



import finders
import teams
import players
import boxscores
import winProb
import pbp

from players import Player
from teams import Team
from boxscores import BoxScore
from finders import GamePlayFinder, PlayerSeasonFinder

BASE_URL = 'http://www.pro-football-reference.com'

# modules/variables to expose
__all__ = [
    'BASE_URL',
    'finders', 'GamePlayFinder', 'PlayerSeasonFinder',
    'boxscores', 'BoxScore',
    'players', 'Player',
    'teams', 'Team',
    'winProb',
    'pbp',
]



import re

import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref

__all__ = [
    'team_names',
    'team_ids',
    'list_teams',
    'Team',
]


@sportsref.decorators.memoized
def team_names(year):
    """Returns a mapping from team ID to full team name for a given season.
    Example of a full team name: "New England Patriots"

    :year: The year of the season in question (as an int).
    :returns: A dictionary with teamID keys and full team name values.
    """
    doc = pq(sportsref.utils.get_html(sportsref.nfl.BASE_URL + '/teams/'))
    active_table = doc('table#teams_active')
    active_df = sportsref.utils.parse_table(active_table)
    inactive_table = doc('table#teams_inactive')
    inactive_df = sportsref.utils.parse_table(inactive_table)
    df = pd.concat((active_df, inactive_df))
    df = df.loc[~df['hasClass_partial_table']]
    ids = df.team_name.str[:3].values
    names = [tr('th a') for tr in active_table('tr').items()]
    names.extend(tr('th a') for tr in inactive_table('tr').items())
    names = filter(None, names)
    names = [lst[0].text_content() for lst in names]
    # combine IDs and team names into pandas series
    series = pd.Series(names, index=ids)
    # create a mask to filter to teams from the given year
    mask = ((df.year_min <= year) & (year <= df.year_max)).values
    # filter, convert to a dict, and return
    return series[mask].to_dict()


@sportsref.decorators.memoized
def team_ids(year):
    """Returns a mapping from team name to team ID for a given season. Inverse
    mapping of team_names. Example of a full team name: "New England Patriots"

    :year: The year of the season in question (as an int).
    :returns: A dictionary with full team name keys and teamID values.
    """
    names = team_names(year)
    return {v: k for k, v in names.iteritems()}


@sportsref.decorators.memoized
def list_teams(year):
    """Returns a list of team IDs for a given season.

    :year: The year of the season in question (as an int).
    :returns: A list of team IDs.
    """
    return team_names(year).keys()


@sportsref.decorators.memoized
class Team:

    def __init__(self, teamID):
        self.teamID = teamID

    def __eq__(self, other):
        return (self.teamID == other.teamID)

    def __hash__(self):
        return hash(self.teamID)

    def __repr__(self):
        return 'Team({})'.format(self.teamID)

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return Team, (self.teamID,)

    @sportsref.decorators.memoized
    def team_year_url(self, yr_str):
        return (sportsref.nfl.BASE_URL +
                '/teams/{}/{}.htm'.format(self.teamID, yr_str))

    @sportsref.decorators.memoized
    def get_main_doc(self):
        relURL = '/teams/{}'.format(self.teamID)
        teamURL = sportsref.nfl.BASE_URL + relURL
        mainDoc = pq(sportsref.utils.get_html(teamURL))
        return mainDoc

    @sportsref.decorators.memoized
    def get_year_doc(self, yr_str):
        return pq(sportsref.utils.get_html(self.team_year_url(yr_str)))

    @sportsref.decorators.memoized
    def name(self):
        """Returns the real name of the franchise given the team ID.

        Examples:
        'nwe' -> 'New England Patriots'
        'sea' -> 'Seattle Seahawks'

        :returns: A string corresponding to the team's full name.
        """
        doc = self.get_main_doc()
        headerwords = doc('div#meta h1')[0].text_content().split()
        lastIdx = headerwords.index('Franchise')
        teamwords = headerwords[:lastIdx]
        return ' '.join(teamwords)

    @sportsref.decorators.memoized
    def roster(self, year):
        """Returns the roster table for the given year.

        :year: The year for which we want the roster; defaults to current year.
        :returns: A DataFrame containing roster information for that year.
        """
        doc = self.get_year_doc('{}_roster'.format(year))
        table = doc('table#games_played_team')
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoized
    def boxscores(self, year):
        """Gets list of BoxScore objects corresponding to the box scores from
        that year.

        :year: The year for which we want the boxscores; defaults to current
        year.
        :returns: np.array of strings representing boxscore IDs.
        """
        doc = self.get_year_doc(year)
        table = doc('table#games')
        df = sportsref.utils.parse_table(table)
        if df.empty:
            return np.array([])
        return df.boxscoreID.values

    # TODO: add functions for OC, DC, PF, PA, W-L, etc.
    # TODO: Also give a function at BoxScore.homeCoach and BoxScore.awayCoach
    # TODO: BoxScore needs a gameNum function to do this?

    @sportsref.decorators.memoized
    def head_coaches_by_game(self, year):
        """Returns head coach data by game.

        :year: An int representing the season in question.
        :returns: An array with an entry per game of the season that the team
        played (including playoffs). Each entry is the head coach's ID for that
        game in the season.
        """
        doc = self.get_year_doc(year)
        coaches = (doc('div#meta p')
                   .filter(lambda i, e: 'Coach:' in e.text_content()))
        coachStr = sportsref.utils.flatten_links(coaches)
        regex = r'(\S+?) \((\d+)-(\d+)-(\d+)\)'
        coachAndTenure = []
        while coachStr:
            m = re.search(regex, coachStr)
            coachID, wins, losses, ties = m.groups()
            nextIndex = m.end(4) + 1
            coachStr = coachStr[nextIndex:]
            tenure = int(wins) + int(losses) + int(ties)
            coachAndTenure.append((coachID, tenure))

        coachIDs = [[cID for _ in xrange(games)]
                    for cID, games in coachAndTenure]
        coachIDs = [cID for sublist in coachIDs for cID in sublist]
        return np.array(coachIDs[::-1])

    @sportsref.decorators.memoized
    def srs(self, year):
        """Returns the SRS (Simple Rating System) for a team in a year.

        :year: The year for the season in question.
        :returns: A float of SRS.
        """
        doc = self.get_year_doc(year)
        srsText = (doc('div#meta p')
                   .filter(lambda i, e: 'SRS' in e.text_content())
                   .text())
        m = re.match(r'SRS\s*?:\s*?(\S+)', srsText)
        if m:
            return float(m.group(1))
        else:
            return np.nan

    @sportsref.decorators.memoized
    def sos(self, year):
        """Returns the SOS (Strength of Schedule) for a team in a year, based
        on SRS.

        :year: The year for the season in question.
        :returns: A float of SOS.
        """
        doc = self.get_year_doc(year)
        sosText = (doc('div#meta p')
                   .filter(lambda i, e: 'SOS' in e.text_content())
                   .text())
        m = re.search(r'SOS\s*?:\s*?(\S+)', sosText)
        if m:
            return float(m.group(1))
        else:
            return np.nan

    @sportsref.decorators.memoized
    def stadium(self, year):
        """Returns the ID for the stadium in which the team played in a given
        year.

        :year: The year in question.
        :returns: A string representing the stadium ID.
        """
        doc = self.get_year_doc(year)
        anchor = (doc('div#meta p')
                  .filter(lambda i, e: 'Stadium' in e.text_content())
                  )('a')
        return sportsref.utils.rel_url_to_id(anchor.attr['href'])

    @sportsref.decorators.memoized
    def team_stats(self, year):
        """Returns a Series (dict-like) of team stats from the team-season
        page.

        :year: Int representing the season.
        :returns: A Series of team stats.
        """
        doc = self.get_year_doc(year)
        table = doc('table#team_stats')
        df = sportsref.utils.parse_table(table)
        return df.ix[df.player_id == 'Team Stats'].iloc[0]

    @sportsref.decorators.memoized
    def opp_stats(self, year):
        """Returns a Series (dict-like) of the team's opponent's stats from the
        team-season page.

        :year: Int representing the season.
        :returns: A Series of team stats.
        """
        doc = self.get_year_doc(year)
        table = doc('table#team_stats')
        df = sportsref.utils.parse_table(table)
        return df.ix[df.player_id == 'Opp. Stats'].iloc[0]

    @sportsref.decorators.memoized
    def passing(self, year):
        doc = self.get_year_doc(year)
        table = doc('table#passing')
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoized
    def rushing_and_receiving(self, year):
        doc = self.get_year_doc(year)
        table = doc('#rushing_and_receiving')
        df = sportsref.utils.parse_table(table)
        return df



import copy
import re

import numpy as np
import pandas as pd

import sportsref


@sportsref.decorators.memoized
def parse_play(details, hm, aw, is_hm, yr):
    """Parse play details from a play-by-play string describing a play; returns
    structured data in a dictionary.

    :details: detail string for the play
    :hm: the ID of the home team (for identifying team w/ possession)
    :aw: the ID of the away team (for identifying team w/ possession)
    :is_hm: bool indicating whether it is a home play (for possession)
    :yr: year of the game
    :returns: dictionary of play attributes; -1 on a play that should be
    skipped
    """
    # if input isn't a string, return -1
    if not isinstance(details, basestring):
        return -1

    p = {}
    p['detail'] = details
    p['home'] = hm
    p['away'] = aw
    p['isHomePlay'] = is_hm

    playerRE = r'\w{0,7}\d{2}'

    # parsing field goal attempts
    shotRE = (r'(?P<shooter>{0}) (?P<isFGM>makes|misses) (?P<isThree>2|3)\-pt'
              r' shot').format(playerRE)
    distRE = r' (?:from (?P<shotDist>\d+) ft|at rim)'
    assistRE = r' \(assist by (?P<assister>{0})\)'.format(playerRE)
    blockRE = r' \(block by (?P<blocker>{0})\)'.format(playerRE)
    shotRE = r'{0}{1}(?:{2}|{3})?'.format(shotRE, distRE, assistRE, blockRE)
    m = re.match(shotRE, details, re.IGNORECASE)
    if m:
        p['isFGA'] = True
        p.update(m.groupdict())
        p['shotDist'] = p['shotDist'] if p['shotDist'] is not None else 0
        p['shotDist'] = int(p['shotDist'])
        p['isFGM'] = p['isFGM'] == 'makes'
        p['isThree'] = p['isThree'] == '3'
        p['isAssist'] = pd.notnull(p.get('assister'))
        p['isBlock'] = pd.notnull(p.get('blocker'))
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        return p

    # parsing jump balls
    jumpRE = (r'Jump ball: (?P<awayJumper>{0}) vs\. (?P<homeJumper>{0})'
              r'(?: \((?P<gainsPos>{0}) gains possession\))?').format(playerRE)
    m = re.match(jumpRE, details, re.IGNORECASE)
    if m:
        p['isJumpBall'] = True
        p.update(m.groupdict())
        return p

    # parsing rebounds
    rebRE = (r'(?P<isOReb>Offensive|Defensive) rebound'
             r' by (?P<rebounder>{0}|Team)').format(playerRE)
    m = re.match(rebRE, details, re.I)
    if m:
        p['isReb'] = True
        p.update(m.groupdict())
        p['isOReb'] = p['isOReb'] == 'Offensive'
        p['isDReb'] = not p['isOReb']
        p['rebTeam'], other = (hm, aw) if is_hm else (aw, hm)
        p['team'] = p['rebTeam'] if p['isOReb'] else other
        p['opp'] = p['rebTeam'] if p['isDReb'] else other
        return p

    # parsing shooting fouls
    shotFoulRE = (r'Shooting(?P<isBlockFoul> block)? foul by (?P<fouler>{0})'
                  r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(shotFoulRE, details, re.I)
    if m:
        p['isShotFoul'] = True
        p.update(m.groupdict())
        p['isBlockFoul'] = p['isBlockFoul'] == ' block'
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        return p

    # parsing free throws
    ftRE = (r'(?P<ftShooter>{}) (?P<isFTM>makes|misses) '
            r'(?P<isTechFT>technical )?(?P<isFlagFT>flagrant )?'
            r'(?P<isClearPathFT>clear path )?free throw'
            r'(?: (?P<ftNum>\d+) of (?P<totFTAtt>\d+))?').format(playerRE)
    m = re.match(ftRE, details, re.I)
    if m:
        p['isFTA'] = True
        p.update(m.groupdict())
        p['isFTM'] = p['isFTM'] == 'makes'
        p['isTechFT'] = p['isTechFT'] == 'technical '
        p['isFlagFT'] = p['isFlagFT'] == 'flagrant '
        p['isClearPathFT'] = p['isClearPathFT'] == 'clear path '
        if p['totFTAtt']:
            p['totFTAtt'] = int(p['totFTAtt'])
        if p['ftNum']:
            p['ftNum'] = int(p['ftNum'])
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        return p

    # parsing substitutions
    subRE = (r'(?P<subIn>{0}) enters the game for '
             r'(?P<subOut>{0})').format(playerRE)
    m = re.match(subRE, details, re.I)
    if m:
        p['isSub'] = True
        p.update(m.groupdict())
        p['subTeam'] = hm if is_hm else aw
        return p

    # parsing turnovers
    toReasons = (r'(?P<toType>[^;]+)(?:; steal by '
                 r'(?P<stealer>{0}))?').format(playerRE)
    toRE = (r'Turnover by (?P<turnoverBy>{}|Team) '
            r'\((?:{})\)').format(playerRE, toReasons)
    m = re.match(toRE, details, re.I)
    if m:
        p['isTO'] = True
        p.update(m.groupdict())
        if p['toType'].lower() == 'offensive foul':
            return -1
        p['isSteal'] = pd.notnull(p['stealer'])
        p['isTravel'] = p['toType'] == 'traveling'
        p['isShotClockViol'] = p['toType'] == 'shot clock'
        p['isOOB'] = p['toType'] == 'step out of bounds'
        p['isThreeSecViol'] = p['toType'] == '3 sec'
        p['isBackCourtViol'] = p['toType'] == 'back court'
        p['isOffGoaltend'] = p['toType'] == 'offensive goaltending'
        p['isDoubleDribble'] = p['toType'] == 'dbl dribble'
        p['isDiscontDribble'] = p['toType'] == 'discontinued dribble'
        p['isCarry'] = p['toType'] == 'palming'
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        return p

    # parsing offensive fouls
    offFoulRE = (r'Offensive(?P<isCharge> charge)? foul '
                 r'by (?P<turnoverBy>{0})'
                 r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(offFoulRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isOffFoul'] = True
        p['isTO'] = True
        p.update(m.groupdict())
        p['isCharge'] = p['isCharge'] == ' charge'
        p['fouler'] = p['turnoverBy']
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        p['foulTeam'] = p['team']
        return p

    # parsing personal fouls
    foulRE = (r'Personal (?P<isTakeFoul>take )?(?P<isBlockFoul>block )?foul by'
              r' (?P<fouler>{0})(?: \(drawn by '
              r'(?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(foulRE, details, re.I)
    if m:
        p['isFoul'] = True
        p.update(m.groupdict())
        p['isTakeFoul'] = p['isTakeFoul'] == 'take '
        p['isBlockFoul'] = p['isBlockFoul'] == 'block '
        p['team'] = aw if is_hm else hm
        p['opp'] = hm if is_hm else aw
        p['foulTeam'] = p['opp']
        return p

    # parsing loose ball fouls
    looseBallRE = (r'Loose ball foul by (?P<fouler>{0})'
                   r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(looseBallRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isLooseBallFoul'] = True
        p.update(m.groupdict())
        p['foulTeam'] = hm if is_hm else aw
        return p

    # parsing away from play fouls
    awayFromBallRE = (r'Away from play foul by (?P<fouler>{0})'
                      r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(awayFromBallRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isAwayFromBallFoul'] = True
        p.update(m.groupdict())
        p['team'] = aw if is_hm else hm
        p['opp'] = hm if is_hm else aw
        p['foulTeam'] = p['opp']
        return p

    # parsing inbound fouls
    inboundRE = (r'Inbound foul by (?P<fouler>{0})'
                 r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(inboundRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isInboundFoul'] = True
        p.update(m.groupdict())
        p['team'] = aw if is_hm else hm
        p['opp'] = hm if is_hm else aw
        p['foulTeam'] = p['opp']
        return p

    # parsing flagrant fouls
    flagrantRE = (r'Flagrant foul type (?P<flagType>1|2) by (?P<fouler>{0})'
                  r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(flagrantRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isFlagrant'] = True
        p.update(m.groupdict())
        p['foulTeam'] = hm if is_hm else aw
        return p

    # parsing clear path fouls
    clearPathRE = (r'Clear path foul by (?P<fouler>{0})'
                   r'(?: \(drawn by (?P<drewFoul>{0})\))?').format(playerRE)
    m = re.match(clearPathRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isClearPathFoul'] = True
        p.update(m.groupdict())
        p['team'] = aw if is_hm else hm
        p['opp'] = hm if is_hm else aw
        p['foulTeam'] = p['opp']
        return p

    # parsing timeouts
    timeoutRE = r'(?P<timeoutTeam>.*?) (?:full )?timeout'
    m = re.match(timeoutRE, details, re.I)
    if m:
        p['isTimeout'] = True
        p.update(m.groupdict())
        p['team'] = hm if is_hm else aw
        p['opp'] = aw if is_hm else hm
        isOfficialTO = p['timeoutTeam'] == 'Official'
        p['timeoutTeam'] = \
            'Official' if isOfficialTO else \
            sportsref.nba.Season(yr).teamNamesToIDs().get(p['team'], p['team'])
        return p

    # parsing technical fouls
    techRE = r'Technical foul by (?P<fouler>{}|Team)'.format(playerRE)
    m = re.match(techRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isTechFoul'] = True
        p.update(m.groupdict())
        p['foulTeam'] = hm if is_hm else aw
        return p

    # parsing ejections
    ejectRE = r'(?P<ejectee>{0}) ejected from game'.format(playerRE)
    m = re.match(ejectRE, details, re.I)
    if m:
        p['isEjection'] = True
        p['ejecteeTeam'] = hm if is_hm else aw
        return p

    # parsing defensive 3 seconds techs
    def3TechRE = (r'(?:Def 3 sec tech foul|Defensive three seconds)'
                  r' by (?P<fouler>{})').format(playerRE)
    m = re.match(def3TechRE, details, re.I)
    if m:
        p['isFoul'] = True
        p['isTechFoul'] = True
        p['isDefThreeSecs'] = True
        p.update(m.groupdict())
        p['foulTeam'] = hm if is_hm else aw
        return p

    # parsing violations
    violRE = (r'Violation by (?P<violator>{0}|Team) '
              r'\((?P<violType>.*)\)').format(playerRE)
    m = re.match(violRE, details, re.I)
    if m:
        p['isViolation'] = True
        p.update(m.groupdict())
        p['violTeam'] = hm if is_hm else aw
        return p

    p['isError'] = True
    return p


def clean_features(df):
    """Fixes up columns of the passed DataFrame, such as casting T/F columns to
    boolean and filling in NaNs for team and opp.

    :df: DataFrame of play-by-play data.
    :returns: Dataframe with cleaned columns.
    """
    df = copy.deepcopy(df)
    # make indicator columns boolean type (and fill in NaNs)
    boolVals = set([True, False, None, np.nan])
    for c in df:
        if set(df[c].unique()[:5]) <= boolVals:
            df[c] = df[c].map(lambda x: x is True)

    # fix free throw columns on technicals
    df.ix[df.isTechFT, ['ftNum', 'totFTAtt']] = 1

    return df



import datetime
import re

import numpy as np
from pyquery import PyQuery as pq

import sportsref

__all__ = [
    'Player',
]


@sportsref.decorators.memoized
class Player:

    def __init__(self, player_id):
        self.player_id = player_id
        self.main_url = (sportsref.nba.BASE_URL +
                         '/players/{0[0]}/{0}.htm').format(self.player_id)

    def __eq__(self, other):
        return self.player_id == other.player_id

    def __hash__(self):
        return hash(self.player_id)

    def __repr__(self):
        return 'Player({})'.format(self.player_id)

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return Player, (self.player_id,)

    @sportsref.decorators.memoized
    def get_doc(self):
        return pq(sportsref.utils.get_html(self.main_url))

    @sportsref.decorators.memoized
    def name(self):
        """Returns the name of the player as a string."""
        doc = self.get_doc()
        return doc('h1[itemprop="name"]').text()

    @sportsref.decorators.memoized
    def age(self, year, month=10, day=1):
        """Returns the age of the player on a given date.

        :year: int representing the year.
        :month: int representing the month (1-12).
        :day: int representing the day within the month (1-31).
        :returns: Age in years as a float.
        """
        doc = self.get_doc()
        date_string = doc('span[itemprop="birthDate"]').attr('data-birth')
        regex = r'(\d{4})\-(\d{2})\-(\d{2})'
        date_args = map(int, re.match(regex, date_string).groups())
        birth_date = datetime.date(*date_args)
        age_date = datetime.date(year=year, month=month, day=day)
        delta = age_date - birth_date
        age = delta.days / 365.
        return age

    @sportsref.decorators.memoized
    def position(self):
        """TODO: Docstring for position.
        :returns: TODO
        """
        raise Exception('not yet implemented - nba.Player.position')

    @sportsref.decorators.memoized
    def height(self):
        """Returns the player's height (in inches).
        :returns: An int representing a player's height in inches.
        """
        doc = self.get_doc()
        raw = doc('span[itemprop="height"]').text()
        try:
            feet, inches = map(int, raw.split('-'))
            return feet * 12 + inches
        except ValueError:
            return np.nan

    @sportsref.decorators.memoized
    def weight(self):
        """Returns the player's weight (in pounds).
        :returns: An int representing a player's weight in pounds.
        """
        doc = self.get_doc()
        raw = doc('span[itemprop="weight"]').text()
        try:
            weight = re.match(r'(\d+)lb', raw).group(1)
            return int(weight)
        except ValueError:
            return np.nan

    @sportsref.decorators.memoized
    def hand(self):
        """Returns the player's handedness.
        :returns: 'L' for left-handed, 'R' for right-handed.
        """
        doc = self.get_doc()
        hand = re.search(r'Shoots:\s*(L|R)', doc.text()).group(1)
        return hand

    @sportsref.decorators.memoized
    def draft_pick(self):
        """Returns when in the draft the player was picked.
        :returns: TODO
        """
        raise Exception('not yet implemented - nba.Player.draft_pick')

    @sportsref.decorators.kind_rpb(include_type=True)
    def _get_stats_table(self, table_id, kind='R'):
        """Gets a stats table from the player page; helper function that does
        the work for per-game, per-100-poss, etc. stats.

        :table_id: the ID of the HTML table.
        :kind: specifies regular season, playoffs, or both. One of 'R', 'P',
        'B'.
        :returns: A DataFrame of stats.
        """
        doc = self.get_doc()
        table_id = 'table#{}{}'.format(
            'playoffs_' if kind == 'P' else '', table_id)
        table = doc(table_id)
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoized
    def stats_per_game(self, kind='R'):
        """Returns a DataFrame of per-game box score stats."""
        return self._get_stats_table('per_game', kind=kind)

    @sportsref.decorators.memoized
    def stats_totals(self, kind='R'):
        """Returns a DataFrame of total box score statistics by season."""
        return self._get_stats_table('totals', kind=kind)

    @sportsref.decorators.memoized
    def stats_per36(self, kind='R'):
        """Returns a DataFrame of per-36-minutes stats."""
        return self._get_stats_table('per_minute', kind=kind)

    @sportsref.decorators.memoized
    def stats_per100(self, kind='R'):
        """Returns a DataFrame of per-100-possession stats."""
        return self._get_stats_table('per_poss', kind=kind)

    @sportsref.decorators.memoized
    def stats_advanced(self, kind='R'):
        """Returns a DataFrame of advanced stats."""
        return self._get_stats_table('advanced', kind=kind)

    @sportsref.decorators.memoized
    def stats_shooting(self, kind='R'):
        """Returns a DataFrame of shooting stats."""
        return self._get_stats_table('shooting', kind=kind)

    @sportsref.decorators.memoized
    def stats_pbp(self, kind='R'):
        """Returns a DataFrame of play-by-play stats."""
        return self._get_stats_table('advanced_pbp', kind=kind)



import re
import datetime

import numpy as np
import pandas as pd
from pyquery import PyQuery as pq

import sportsref

__all__ = [
    'BoxScore',
]


@sportsref.decorators.memoized
class BoxScore:

    def __init__(self, boxscoreID):
        self.boxscoreID = boxscoreID

    def __eq__(self, other):
        return self.boxscoreID == other.boxscoreID

    def __hash__(self):
        return hash(self.boxscoreID)

    def __repr__(self):
        return 'BoxScore({})'.format(self.boxscoreID)

    def __str__(self):
        return '{} Week {}: {} @ {}'.format(
            self.season(), self.week(), self.away(), self.home()
        )

    def __reduce__(self):
        return BoxScore, (self.boxscoreID,)

    @sportsref.decorators.memoized
    def get_doc(self):
        url = (sportsref.nfl.BASE_URL +
               '/boxscores/{}.htm'.format(self.boxscoreID))
        doc = pq(sportsref.utils.get_html(url))
        return doc

    @sportsref.decorators.memoized
    def date(self):
        """Returns the date of the game. See Python datetime.date documentation
        for more.
        :returns: A datetime.date object with year, month, and day attributes.
        """
        match = re.match(r'(\d{4})(\d{2})(\d{2})', self.boxscoreID)
        year, month, day = map(int, match.groups())
        return datetime.date(year=year, month=month, day=day)

    @sportsref.decorators.memoized
    def weekday(self):
        """Returns the day of the week on which the game occurred.
        :returns: String representation of the day of the week for the game.

        """
        days = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday',
                'Saturday', 'Sunday']
        date = self.date()
        wd = date.weekday()
        return days[wd]

    @sportsref.decorators.memoized
    def home(self):
        """Returns home team ID.
        :returns: 3-character string representing home team's ID.
        """
        doc = self.get_doc()
        table = doc('table.linescore')
        relURL = table('tr').eq(1)('a').eq(2).attr['href']
        home = sportsref.utils.rel_url_to_id(relURL)
        return home

    @sportsref.decorators.memoized
    def away(self):
        """Returns away team ID.
        :returns: 3-character string representing away team's ID.
        """
        doc = self.get_doc()
        table = doc('table.linescore')
        relURL = table('tr').eq(2)('a').eq(2).attr['href']
        away = sportsref.utils.rel_url_to_id(relURL)
        return away

    @sportsref.decorators.memoized
    def home_score(self):
        """Returns score of the home team.
        :returns: int of the home score.
        """
        doc = self.get_doc()
        table = doc('table.linescore')
        home_score = table('tr').eq(1)('td')[-1].text_content()
        return int(home_score)

    @sportsref.decorators.memoized
    def away_score(self):
        """Returns score of the away team.
        :returns: int of the away score.
        """
        doc = self.get_doc()
        table = doc('table.linescore')
        away_score = table('tr').eq(2)('td')[-1].text_content()
        return int(away_score)

    @sportsref.decorators.memoized
    def winner(self):
        """Returns the team ID of the winning team. Returns NaN if a tie."""
        hmScore = self.home_score()
        awScore = self.away_score()
        if hmScore > awScore:
            return self.home()
        elif hmScore < awScore:
            return self.away()
        else:
            return np.nan

    @sportsref.decorators.memoized
    def week(self):
        """Returns the week in which this game took place. 18 is WC round, 19
        is Div round, 20 is CC round, 21 is SB.
        :returns: Integer from 1 to 21.
        """
        doc = self.get_doc()
        rawTxt = doc('div#page_content table').eq(0)('tr td').eq(0).text()
        match = re.search(r'Week (\d+)', rawTxt)
        if match:
            return int(match.group(1))
        else:
            return 21  # super bowl is week 21

    @sportsref.decorators.memoized
    def season(self):
        """
        Returns the year ID of the season in which this game took place.
        Useful for week 17 January games.

        :returns: An int representing the year of the season.
        """
        doc = self.get_doc()
        rawTxt = doc('div#page_content table').eq(0)('tr td').eq(0).text()
        match = re.search(r'Week \d+ (\d{4})', rawTxt)
        if match:
            return int(match.group(1))
        else:
            # else, it's the super bowl; super bowl happens in calendar year
            # after the season's year
            return self.date().year - 1

    @sportsref.decorators.memoized
    def starters(self):
        """Returns a DataFrame where each row is an entry in the starters table
        from PFR.

        The columns are:
        * player_id - the PFR player ID for the player (note that this column is
        not necessarily all unique; that is, one player can be a starter in
        multiple positions, in theory).
        * playerName - the listed name of the player; this too is not
        necessarily unique.
        * position - the position at which the player started for their team.
        * team - the team for which the player started.
        * home - True if the player's team was at home, False if they were away
        * offense - True if the player is starting on an offensive position,
        False if defense.

        :returns: A pandas DataFrame. See the description for details.
        """
        doc = self.get_doc()
        a = doc('table#vis_starters')
        h = doc('table#home_starters')
        data = []
        for h, table in enumerate((a, h)):
            team = self.home() if h else self.away()
            for i, row in enumerate(table('tbody tr').items()):
                datum = {}
                datum['player_id'] = sportsref.utils.rel_url_to_id(
                    row('a')[0].attrib['href']
                )
                datum['playerName'] = row('th').text()
                datum['position'] = row('td').text()
                datum['team'] = team
                datum['home'] = (h == 1)
                datum['offense'] = (i <= 10)
                data.append(datum)
        return pd.DataFrame(data)

    @sportsref.decorators.memoized
    def line(self):
        doc = self.get_doc()
        table = doc('table#game_info')
        giTable = sportsref.utils.parse_info_table(table)
        line_text = giTable.get('vegas_line', None)
        if line_text is None:
            return np.nan
        m = re.match(r'(.+?) ([\-\.\d]+)$', line_text)
        if m:
            favorite, line = m.groups()
            line = float(line)
            # give in terms of the home team
            year = self.season()
            if favorite != sportsref.nfl.teams.team_names(year)[self.home()]:
                line = -line
        else:
            line = 0
        return line

    @sportsref.decorators.memoized
    def surface(self):
        """The playing surface on which the game was played.

        :returns: string representing the type of surface. Returns np.nan if
        not avaiable.
        """
        doc = self.get_doc()
        table = doc('table#game_info')
        giTable = sportsref.utils.parse_info_table(table)
        return giTable.get('surface', np.nan)

    @sportsref.decorators.memoized
    def over_under(self):
        """
        Returns the over/under for the game as a float, or np.nan if not
        available.
        """
        doc = self.get_doc()
        table = doc('table#game_info')
        giTable = sportsref.utils.parse_info_table(table)
        if 'over_under' in giTable:
            ou = giTable['over_under']
            return float(ou.split()[0])
        else:
            return np.nan

    @sportsref.decorators.memoized
    def coin_toss(self):
        """Gets information relating to the opening coin toss.

        Keys are:
        * wonToss - contains the ID of the team that won the toss
        * deferred - bool whether the team that won the toss deferred it

        :returns: Dictionary of coin toss-related info.
        """
        doc = self.get_doc()
        table = doc('table#game_info')
        giTable = sportsref.utils.parse_info_table(table)
        if 'Won Toss' in giTable:
            # TODO: finish coinToss function
            pass
        else:
            return np.nan

    @sportsref.decorators.memoized
    def weather(self):
        """Returns a dictionary of weather-related info.

        Keys of the returned dict:
        * temp
        * windChill
        * relHumidity
        * windMPH

        :returns: Dict of weather data.
        """
        doc = self.get_doc()
        table = doc('table#game_info')
        giTable = sportsref.utils.parse_info_table(table)
        if 'weather' in giTable:
            regex = (
                r'(?:(?P<temp>\-?\d+) degrees )?'
                r'(?:relative humidity (?P<relHumidity>\d+)%, )?'
                r'(?:wind (?P<windMPH>\d+) mph, )?'
                r'(?:wind chill (?P<windChill>\-?\d+))?'
            )
            m = re.match(regex, giTable['weather'])
            d = m.groupdict()

            # cast values to int
            for k in d:
                try:
                    d[k] = int(d[k])
                except TypeError:
                    pass

            # one-off fixes
            d['windChill'] = (d['windChill'] if pd.notnull(d['windChill'])
                              else d['temp'])
            d['windMPH'] = d['windMPH'] if pd.notnull(d['windMPH']) else 0
            return d
        else:
            # no weather found, because it's a dome
            # TODO: what's relative humidity in a dome?
            return {
                'temp': 70, 'windChill': 70, 'relHumidity': None, 'windMPH': 0
            }

    @sportsref.decorators.memoized
    def pbp(self):
        """Returns a dataframe of the play-by-play data from the game.

        :returns: pandas DataFrame of play-by-play. Similar to GPF.
        """
        doc = self.get_doc()
        table = doc('table#pbp')
        df = sportsref.utils.parse_table(table)
        # make the following features conveniently available on each row
        df['boxscoreID'] = self.boxscoreID
        df['home'] = self.home()
        df['away'] = self.away()
        df['season'] = self.season()
        df['week'] = self.week()
        feats = sportsref.nfl.pbp.expand_details(df)

        # add team and opp columns by iterating through rows
        df = sportsref.nfl.pbp.add_team_columns(feats)
        # add WPA column (requires diff, can't be done row-wise)
        df['home_wpa'] = df.home_wp.diff()
        # lag score columns, fill in 0-0 to start
        for col in ('home_wp', 'pbp_score_hm', 'pbp_score_aw'):
            if col in df.columns:
                df[col] = df[col].shift(1)
        df.ix[0, ['pbp_score_hm', 'pbp_score_aw']] = 0
        # fill in WP NaN's
        df.home_wp.fillna(method='ffill', inplace=True)
        # fix first play border after diffing/shifting for WP and WPA
        firstPlaysOfGame = df[df.secsElapsed == 0].index
        line = self.line()
        for i in firstPlaysOfGame:
            initwp = sportsref.nfl.winProb.initialWinProb(line)
            df.ix[i, 'home_wp'] = initwp
            df.ix[i, 'home_wpa'] = df.ix[i + 1, 'home_wp'] - initwp
        # fix last play border after diffing/shifting for WP and WPA
        lastPlayIdx = df.iloc[-1].name
        lastPlayWP = df.ix[lastPlayIdx, 'home_wp']
        # if a tie, final WP is 50%; otherwise, determined by winner
        winner = self.winner()
        finalWP = 50. if pd.isnull(winner) else (winner == self.home()) * 100.
        df.ix[lastPlayIdx, 'home_wpa'] = finalWP - lastPlayWP
        # fix WPA for timeouts and plays after timeouts
        timeouts = df[df.isTimeout].index
        for to in timeouts:
            df.ix[to, 'home_wpa'] = 0.
            if to + 2 in df.index:
                wpa = df.ix[to + 2, 'home_wp'] - df.ix[to + 1, 'home_wp']
            else:
                wpa = finalWP - df.ix[to + 1, 'home_wp']
            df.ix[to + 1, 'home_wpa'] = wpa
        # add team-related features to DataFrame
        df = df.apply(sportsref.nfl.pbp.add_team_features, axis=1)
        # fill distToGoal NaN's
        df['distToGoal'] = np.where(df.isKickoff, 65, df.distToGoal)
        df.distToGoal.fillna(method='bfill', inplace=True)
        df.distToGoal.fillna(method='ffill', inplace=True)  # for last play

        return df

    @sportsref.decorators.memoized
    def ref_info(self):
        """Gets a dictionary of ref positions and the ref IDs of the refs for
        that game.

        :returns: A dictionary of ref positions and IDs.
        """
        doc = self.get_doc()
        table = doc('table#officials')
        return sportsref.utils.parse_info_table(table)

    @sportsref.decorators.memoized
    def player_stats(self):
        """Gets the stats for offense, defense, returning, and kicking of
        individual players in the game.
        :returns: A DataFrame containing individual player stats.
        """
        doc = self.get_doc()
        tableIDs = ('player_offense', 'player_defense', 'returns', 'kicking')
        dfs = []
        for tID in tableIDs:
            table = doc('#{}'.format(tID))
            dfs.append(sportsref.utils.parse_table(table))
        df = pd.concat(dfs, ignore_index=True)
        df = df.reset_index(drop=True)
        df['team'] = df['team'].str.lower()
        return df



from setuptools import setup, find_packages

setup(name='sportsref',
      version='0.7.3',
      description='Scraping data from sports-reference.com and related sites',
      url='https://github.com/mdgoldberg/sportsref',
      author='Matt Goldberg',
      author_email='mgoldberg@college.harvard.edu',
      license='Apache License 2.0',
      packages=find_packages(),
      install_requires=[
          'appdirs',
          'numexpr',
          'numpy',
          'pandas',
          'pyquery',
          'scipy',
          'selenium',
      ]
      )



import datetime
import re
import urlparse

import numpy as np
from pyquery import PyQuery as pq

import sportsref

__all__ = [
    'Player',
]


@sportsref.decorators.memoized
class Player:

    def __init__(self, player_id):
        self.player_id = player_id
        self.mainURL = (sportsref.nfl.BASE_URL +
                        '/players/{0[0]}/{0}.htm').format(self.player_id)

    def __eq__(self, other):
        return self.player_id == other.player_id

    def __hash__(self):
        return hash(self.player_id)

    def __repr__(self):
        return 'Player({})'.format(self.player_id)

    def __str__(self):
        return self.name()

    def __reduce__(self):
        return Player, (self.player_id,)

    def _subpage_url(self, page, year=None):
        # if no year, return career version
        if year is None:
            return urlparse.urljoin(
                self.mainURL, '{}/{}/'.format(self.player_id, page)
            )
        # otherwise, return URL for a given year
        else:
            return urlparse.urljoin(
                self.mainURL, '{}/{}/{}/'.format(self.player_id, page, year)
            )

    @sportsref.decorators.memoized
    def get_doc(self):
        doc = pq(sportsref.utils.get_html(self.mainURL))
        return doc

    @sportsref.decorators.memoized
    def name(self):
        doc = self.get_doc()
        name = doc('div#meta h1:first').text()
        return name

    @sportsref.decorators.memoized
    def age(self, year, month=9, day=1):
        doc = self.get_doc()
        span = doc('div#meta span#necro-birth')
        birthstring = span.attr('data-birth')
        try:
            dateargs = re.match(r'(\d{4})\-(\d{2})\-(\d{2})',
                                birthstring).groups()
            dateargs = map(int, dateargs)
            birthDate = datetime.date(*dateargs)
            delta = datetime.date(year=year, month=month, day=day) - birthDate
            age = delta.days / 365.
            return age
        except Exception:
            return np.nan

    @sportsref.decorators.memoized
    def position(self):
        doc = self.get_doc()
        rawText = (doc('div#meta p')
                   .filter(lambda i, e: 'Position' in e.text_content())
                   .text())
        rawPos = re.search(r'Position\W*(\S+)', rawText, re.I).group(1)
        allPositions = rawPos.split('-')
        # right now, returning just the primary position for those with
        # multiple positions
        return allPositions[0]

    @sportsref.decorators.memoized
    def height(self):
        doc = self.get_doc()
        rawText = doc('div#meta p span[itemprop="height"]').text()
        try:
            feet, inches = map(int, rawText.split('-'))
            return feet * 12 + inches
        except ValueError:
            return np.nan

    @sportsref.decorators.memoized
    def weight(self):
        doc = self.get_doc()
        rawText = doc('div#meta p span[itemprop="weight"]').text()
        try:
            weight = re.match(r'(\d+)lb', rawText, re.I).group(1)
            return int(weight)
        except AttributeError:
            return np.nan

    @sportsref.decorators.memoized
    def hand(self):
        doc = self.get_doc()
        try:
            rawText = (doc('div#meta p')
                       .filter(lambda i, e: 'Throws' in e.text_content())
                       .text())
            rawHand = re.search(r'Throws\W+(\S+)', rawText, re.I).group(1)
        except AttributeError:
            return np.nan
        return rawHand[0]  # 'L' or 'R'

    @sportsref.decorators.memoized
    def current_team(self):
        doc = self.get_doc()
        team = (doc('div#meta p')
                .filter(lambda i, e: 'Team' in e.text_content()))
        text = sportsref.utils.flatten_links(team)
        try:
            m = re.match(r'Team: (\w{3})', text)
            return m.group(1)
        except Exception:
            return np.nan

    @sportsref.decorators.memoized
    def draft_pick(self):
        doc = self.get_doc()
        rawDraft = (doc('div#meta p')
                    .filter(lambda i, e: 'Draft' in e.text_content())
                    .text())
        m = re.search(r'Draft.*? round \((\d+).*?overall\)', rawDraft, re.I)
        # if not drafted or taken in supplemental draft, return NaN
        if m is None or 'Supplemental' in rawDraft:
            return np.nan
        else:
            return int(m.group(1))

    @sportsref.decorators.memoized
    def draft_class(self):
        doc = self.get_doc()
        rawDraft = (doc('div#meta p')
                    .filter(lambda i, e: 'Draft' in e.text_content())
                    .text())
        m = re.search(r'Draft.*?of the (\d{4}) NFL', rawDraft, re.I)
        if not m:
            return np.nan
        else:
            return int(m.group(1))

    @sportsref.decorators.memoized
    def draft_team(self):
        doc = self.get_doc()
        rawDraft = (doc('div#meta p')
                    .filter(lambda i, e: 'Draft' in e.text_content()))
        try:
            draftStr = sportsref.utils.flatten_links(rawDraft)
            m = re.search(r'Draft\W+(\w+)', draftStr)
            return m.group(1)
        except Exception:
            return np.nan

    @sportsref.decorators.memoized
    def college(self):
        doc = self.get_doc()
        rawText = (doc('div#meta p')
                   .filter(lambda i, e: 'College' in e.text_content()))
        cleanedText = sportsref.utils.flatten_links(rawText)
        college = re.search(r'College:\s*(\S+)', cleanedText).group(1)
        return college

    @sportsref.decorators.memoized
    def high_school(self):
        doc = self.get_doc()
        rawText = (doc('div#meta p')
                   .filter(lambda i, e: 'High School' in e.text_content()))
        cleanedText = sportsref.utils.flatten_links(rawText)
        hs = re.search(r'High School:\s*(\S+)', cleanedText).group(1)
        return hs

    @sportsref.decorators.memoized
    @sportsref.decorators.kind_rpb(include_type=True)
    def gamelog(self, year=None, kind='R'):
        """Gets the career gamelog of the given player.
        :kind: One of 'R', 'P', or 'B' (for regular season, playoffs, or both).
        Case-insensitive; defaults to 'R'.
        :year: The year for which the gamelog should be returned; if None,
        return entire career gamelog. Defaults to None.
        :returns: A DataFrame with the player's career gamelog.
        """
        url = self._subpage_url('gamelog', None)  # year is filtered later
        doc = pq(sportsref.utils.get_html(url))
        table = doc('#stats') if kind == 'R' else doc('#stats_playoffs')
        df = sportsref.utils.parse_table(table)
        if year is not None:
            df = df.query('year == @year').reset_index(drop=True)
        return df

    @sportsref.decorators.memoized
    @sportsref.decorators.kind_rpb(include_type=True)
    def passing(self, kind='R'):
        """Gets yearly passing stats for the player.

        :kind: One of 'R', 'P', or 'B'. Case-insensitive; defaults to 'R'.
        :returns: Pandas DataFrame with passing stats.
        """
        doc = self.get_doc()
        table = doc('#passing') if kind == 'R' else doc('#passing_playoffs')
        df = sportsref.utils.parse_table(table)
        return df

    @sportsref.decorators.memoized
    @sportsref.decorators.kind_rpb(include_type=True)
    def rushing_and_receiving(self, kind='R'):
        """Gets yearly rushing/receiving stats for the player.

        :kind: One of 'R', 'P', or 'B'. Case-insensitive; defaults to 'R'.
        :returns: Pandas DataFrame with rushing/receiving stats.
        """
        doc = self.get_doc()
        table = (doc('#rushing_and_receiving') if kind == 'R'
                 else doc('#rushing_and_receiving_playoffs'))
        if not table:
            table = (doc('#receiving_and_rushing') if kind == 'R'
                     else doc('#receiving_and_rushing_playoffs'))
        df = sportsref.utils.parse_table(table)
        return df

    def _plays(self, year, play_type):
        """Returns a DataFrame of plays for a given year for a given play type
        (like rushing, receiving, or passing).

        :year: The year for the season.
        :play_type: A type of play for which there are plays (as of this
        writing, either "passing", "rushing", or "receiving")
        :returns: A DataFrame of plays, each row is a play. Returns None if
        there were no such plays in that year.
        """
        url = self._subpage_url('{}-plays'.format(play_type), year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc('table#all_plays')
        if table:
            plays = sportsref.nfl.pbp.expand_details(
                sportsref.utils.parse_table(table), detailCol='description'
            )
            return plays
        else:
            return None

    @sportsref.decorators.memoized
    def passing_plays(self, year):
        """Returns a pbp DataFrame of a player's passing plays in a season.

        :year: The year for the season.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, 'passing')

    @sportsref.decorators.memoized
    def rushing_plays(self, year):
        """Returns a pbp DataFrame of a player's rushing plays in a season.

        :year: The year for the season.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, 'rushing')

    @sportsref.decorators.memoized
    def receiving_plays(self, year):
        """Returns a pbp DataFrame of a player's receiving plays in a season.

        :year: The year for the season.
        :returns: A DataFrame of stats, each row is a play.
        """
        return self._plays(year, 'receiving')

    @sportsref.decorators.memoized
    def splits(self, year=None):
        """Returns a DataFrame of splits data for a player-year.

        :year: The year for the season in question. If None, returns career
        splits.
        :returns: A DataFrame of splits data.
        """
        # get the table
        url = self._subpage_url('splits', year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc('table#stats')
        df = sportsref.utils.parse_table(table)
        # cleaning the data
        if not df.empty:
            df.split_id.fillna(method='ffill', inplace=True)
            df.set_index(['split_id', 'split_value'], inplace=True)
        return df

    @sportsref.decorators.memoized
    def advanced_splits(self, year=None):
        """Returns a DataFrame of advanced splits data for a player-year.

        :year: The year for the season in question. If None, returns career
        advanced splits.
        :returns: A DataFrame of advanced splits data.
        """
        # get the table
        url = self._subpage_url('splits', year)
        doc = pq(sportsref.utils.get_html(url))
        table = doc('table#advanced_splits')
        df = sportsref.utils.parse_table(table)
        # cleaning the data
        if not df.empty:
            df.split_type.fillna(method='ffill', inplace=True)
            df.set_index(['split_type', 'split_value'], inplace=True)
        return df
